/**
 * The Max API module to interact and communicate with Max from within the Node context.
 * @module max-api
 * @example
 * // To make this API module available simply require max-api
 * const maxAPI = require("max-api");
 */


const { EventEmitter } = require("events");
const loSet = require("lodash.set");

const maxAPI = require("../api");

const { API_MESSAGES, ERROR_CODES } = require("../api/constants");
const { MESSAGE_TYPES, POST_LEVELS } = require("./constants");

const inletEvents = new EventEmitter();

const makeError = (message, code) => {
	const err = new Error(message);
	err.code = code;
	return err;
};

const isPostLevel = (arg) => arg === POST_LEVELS.INFO || arg === POST_LEVELS.WARN || arg === POST_LEVELS.ERROR;

/**
 * Set a handler/callback function for the given message
 *
 * @memberof module:max-api
 * @name addHandler
 * @kind function
 * @param {module:max-api.MessageIdentifier} msg The message identifier to set the handler for
 * @param {module:max-api.MessageHandler} handler The message handler to add
 */
const addHandler = (msg, handler) => {
	inletEvents.addListener(msg, handler);
};

/**
 * Bulk register a set of handlers provided in an object fashion
 *
 * @memberof module:max-api
 * @name addHandlers
 * @kind function
 * @param {Object} handlers The handler objects
 * @example
 *
 * const maxAPI = require("max-api");
 *
 * const handlers = {
 *   [MESSAGE_TYPES.BANG]: () => {
 *     console.log("got a bang");
 *   },
 *   [MESSAGE_TYPES.NUMBER]: (num) => {
 *   },
 *   my_message: () => {
 *     console.log("got my_message");
 *   },
 *   my_message_with_args: (arg1, arg2) => {
 *     console.log("got my arged message: ${arg1}, ${arg2} ");
 *   }
 *   [MESSAGE_TYPES.ALL]: (handled, ...args) => {
 *     console.log("This will be called for ALL messages");
 *     console.log(`The following inlet event was ${!handled ? "not " : "" }handled`);
 *     console.log(args);
 *   }
 * };
 *
 * maxAPI.addHandlers(handlers);
 */
const addHandlers = (handlers) => {
	const msgIds = Object.keys(handlers);
	for (let i = 0, il = msgIds.length; i < il; i++) {
		const id = msgIds[i];
		addHandler(id, handlers[id]);
	}
};

/**
 * Set a handler/callback function for the given message
 * @memberof module:max-api
 * @name removeHandler
 * @kind function
 * @param {module:max-api.MessageIdentifier} msg The message identifier to remove the handler for
 * @param {module:max-api.MessageHandler} handler The message handler to remove
 */
const removeHandler = (msg, handler) => {
	inletEvents.removeListener(msg, handler);
};

/**
 * Remove all inlet handlers for the given MessageIdentifier. If no identifer is provided
 * this function call removes *all* inlet handlers for *all* messages
 *
 * @memberof module:max-api
 * @name removeHandlers
 * @kind function
 * @param {module:max-api.MessageIdentifier} [msg] - The message identifier to remove the handler for
 */
const removeHandlers = (msg) => {
	if (msg) {
		inletEvents.removeAllListeners(msg);
	} else {
		inletEvents.removeAllListeners();
	}
};

/**
 * Outlets the given value of the object's outlet in Max
 *
 * @memberof module:max-api
 * @name outlet
 * @kind function
 * @param {...module:max-api.Anything} anything The value to output
 *
 * @returns {Promise.<null, Error>} Returns null if resolved or an Error if rejected
 */
const outlet = async (...args) => {

	if (!args || !args.length) throw makeError("Missing values for outlet call.", ERROR_CODES.INVALID_PARAMETERS);

	await maxAPI.request(API_MESSAGES.OUTLET, {
		value: args.length === 1 ? args[0] : args
	});

	return;
};

/**
 * Sends a bang out of the object's outlet
 *
 * @memberof module:max-api
 * @name outletBang
 * @kind function
 * @returns {Promise.<null, Error>} Returns null if resolved or an Error if rejected
 */
const outletBang = async () => {
	await maxAPI.request(API_MESSAGES.OUTLET, {
		value: MESSAGE_TYPES.BANG
	});
	return;
};

/**
 * Print the given value to the Max console
 *
 * @memberof module:max-api
 * @name post
 * @kind function
 * @param {...module:max-api.Anything} anything The value to post
 * @param {module:max-api.POST_LEVELS} level=POST_LEVELS.INFO The log level of the post
 *
 * @returns {Promise.<null, Error>} Returns null if resolved or an Error if rejected
 */
const post = async (...message) => {

	if (!message || !message.length) throw makeError("Missing values for post call.", ERROR_CODES.INVALID_PARAMETERS);

	let level = POST_LEVELS.INFO;

	// check if we got a POST_LEVEL as last attribute
	if (message.length > 1 && isPostLevel(message.slice(-1)[0])) level = message.pop();

	await maxAPI.request(API_MESSAGES.POST, {
		level,
		value: message.length === 1 ? message[0] : message
	});
	return;
};

/**
 * Access the contents of a dictionary in Max
 *
 * @memberof module:max-api
 * @name getDict
 * @kind function
 * @param {module:max-api.DictIdentifier} id The identifier of the dictionary
 *
 * @returns {Promise.<module:max-api.Dict, Error>} Return the dictionary if resolved or an Error if rejected
 */
const getDict = async (id) => {

	if (!id) throw makeError("Missing name for getDict request. Please make sure you provide a (valid) name.", ERROR_CODES.INVALID_PARAMETERS);

	const payload = await maxAPI.request(API_MESSAGES.GET_DICT, {
		name: id
	});
	if (!payload || !payload.content) throw makeError("Invalid response for getDict request. Please make sure the requested dict exists.", ERROR_CODES.INVALID_RESPONSE);
	return payload.content;
};

/**
 * Overrides the *entire* content of a dictionary in Max
 *
 * @memberof module:max-api
 * @name setDict
 * @kind function
 * @param {module:max-api.DictIdentifier} id The identifier of the dictionary
 * @param {module:max-api.Dict} content The new content of the dictionary
 *
 * @returns {Promise.<module:max-api.Dict, Error>} Return the updated dictionary if resolved or an Error if rejected
 */
const setDict = async (id, content) => {
	if (!id) throw makeError("Missing name for setDict request. Please make sure you provide a (valid) name.", ERROR_CODES.INVALID_PARAMETERS);
	if (!content) throw makeError("Missing value for setDict request. A dict cannot be set to null.", ERROR_CODES.INVALID_PARAMETERS);

	const payload = await maxAPI.request(API_MESSAGES.SET_DICT, {
		name: id,
		merge: false,
		content
	});
	if (!payload || !payload.content) throw makeError("Invalid response for setDict request. Please make sure the requested dict exists.", ERROR_CODES.INVALID_RESPONSE);
	return payload.content;
};

/**
 * Updates the content of a dictionary in Max at the given path with the given value
 *
 * @memberof module:max-api
 * @name updateDict
 * @kind function
 * @param {module:max-api.DictIdentifier} id The identifier of the dictionary
 * @param {module:max-api.DictPath} path The path of the value change within the dictionary
 * @param {module:max-api.Anything} value The value to set at this path
 *
 * @returns {Promise.<module:max-api.Dict, Error>} Return the updated dictionary if resolved or an Error if rejected
 */
const updateDict = async (id, updatePath, updateValue) => {
	if (!id) throw makeError("Missing name for setDict request. Please make sure you provide a (valid) name.", ERROR_CODES.INVALID_PARAMETERS);
	if (!updatePath) throw makeError("Missing path value for updateDict request.", ERROR_CODES.INVALID_PARAMETERS);

	const dict = await getDict(id);
	const newDict = loSet(dict, updatePath, updateValue);
	await setDict(id, newDict);

	return newDict;
};

maxAPI.on("inlet", async (inletValue) => {

	let selector;
	let args = [];

	if (Array.isArray(inletValue)) {

		[selector, ...args] = inletValue;

		// Is the first element of the list is a number, then prepend
		// the "list" selector
		if (typeof selector === "number") {
			args.unshift(selector);
			selector = MESSAGE_TYPES.LIST;
		}

	} else if (typeof inletValue === "number") {

		args = [inletValue];
		selector = MESSAGE_TYPES.NUMBER;

	} else if (typeof inletValue === "object") {

		args = [inletValue];
		selector = MESSAGE_TYPES.DICT;

	} else if (inletValue === MESSAGE_TYPES.BANG) {

		selector = MESSAGE_TYPES.BANG;
		args = [];

	} else {

		// generic message handler
		selector = inletValue;
		args = [];

	}

	const selectorCount = inletEvents.listenerCount(selector);
	const allCount = inletEvents.listenerCount(MESSAGE_TYPES.ALL);

	if (selectorCount > 0) inletEvents.emit(selector, ...args);
	if (allCount > 0) inletEvents.emit(MESSAGE_TYPES.ALL, selectorCount > 0, selector, ...args);

	if (!(allCount || selectorCount)) {
		try {
			await post(`Unhandled Message: ${JSON.stringify(inletValue)}`, POST_LEVELS.WARN);
		} catch (err) {
			// Simply ignore at this point as it's just a info headsup
		}
	}
});


// Allow the user to register a custom clean-up/shutdown hook
let registeredShutdownFct = null;
const __n4m_getRegisteredShutdownHook = () => {
	return registeredShutdownFct;
};

/**
 * Registers a shutdown function that will be called before the process exits. Use this to perform any
 * clean up tasks
 *
 * @memberof module:max-api
 * @name registerShutdownHook
 * @kind function
 * @param {function} fct The function to call before the process exits
 */
const registerShutdownHook = (fct) => {
	if (!fct || !(fct instanceof Function)) throw makeError("Shutdownhook registration is only possible with a valid function.", ERROR_CODES.INVALID_PARAMETERS);
	registeredShutdownFct = fct;
};


module.exports = {
	// CONSTANTS
	MESSAGE_TYPES,
	POST_LEVELS,

	addHandler,
	addHandlers,
	getDict,
	outlet,
	outletBang,
	post,
	registerShutdownHook,
	removeHandler,
	removeHandlers,
	setDict,
	updateDict,

	__n4m_getRegisteredShutdownHook

};
